"""A Model definition of Vault resources"""

#  Ensure every defined model is loaded I guess
import aomi.model.generic
import aomi.model.auth
import aomi.model.aws
import aomi.model.resource
import aomi.model.backend

from aomi.model.context import Context
